from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class SignalItem(object):
    """
    basic hdl signal used to design circuits
    """

    def __init__(self, name, dtype, defaultVal=None, virtualOnly=False):
        """
        :param name: name for better orientation in netlists (used only in serialization)
        :param dtype: data type of this signal
        :param defaultVal: value for initialization
        :param virtualOnly: flag indicates that this assignments is only virtual and should not be added into
            netlist, because it is only for internal notation
        """
        self.name = name
        self._dtype = dtype
        self.virtualOnly = virtualOnly
        if defaultVal is None:
            defaultVal = dtype.fromPy(None)
        self.defaultVal = defaultVal
        self._setDefValue()

    def _setDefValue(self):
        v = self.defaultVal
        if isinstance(v, RtlSignalBase):
            v = v.staticEval()

        self._val = v.clone()
        self._oldVal = self._val.clone()
        self._oldVal.vldMask = 0

    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer, DebugTmpVarStack
        tmpVars = DebugTmpVarStack()
        ctx = VhdlSerializer.getBaseContext()
        ctx.createTmpVarFn = tmpVars.createTmpVarFn
        s = VhdlSerializer.SignalItem(self, ctx)
        return "%s%s" % (tmpVars.serialize(), s)



VHLD_KEYWORDS = [
    "abs", "access", "across", "after", "alias", "all", "and", "architecture", "array",
    "assert", "attribute", "begin", "block", "body", "break", "bugger", "bus", "case",
    "component", "configuration", "constant", "disconnect", "downto", "end", "entity",
    "else", "elsif", "exit", "file", "for", "function", "generate", "generic", "group",
    "guarded", "if", "impure", "in", "inertial", "inout", "is", "label", "library", "limit",
    "linkage", "literal", "loop", "map", "mod", "nand", "nature", "new", "next", "noise",
    "nor", "not", "null", "of", "on", "open", "or", "others", "out", "package", "port",
    "postponed", "process", "procedure", "procedural", "pure", "quantity", "range",
    "reverse_range", "reject", "rem", "record", "reference", "register", "report", "return",
    "rol", "ror", "select", "severity", "shared", "signal", "sla", "sll", "spectrum", "sra",
    "srl", "subnature", "subtype", "terminal", "then", "through", "to", "tolerance", "transport",
    "type", "unaffected", "units", "until", "use", "variable", "wait", "with", "when", "while",
    "xnor", "xor"]



from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.boolean import Boolean
from hwt.hdlObjects.types.defs import INT
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.simulator.types.simInt import SIM_INT


def convertSimBits__val(self, sigOrVal, toType):
    if isinstance(toType, Boolean):
        return sigOrVal._eq_val(self.getValueCls().fromPy(1, self))

    elif isinstance(toType, Bits):
        if self._widthVal == toType._widthVal:
            return sigOrVal._convSign(toType.signed)

    elif toType == INT or toType == SIM_INT:
        if self.signed:
            raise NotImplementedError()
        else:
            fullMask = self._allMask
            return INT.getValueCls()(sigOrVal.val,
                                     INT,
                                     sigOrVal.vldMask == fullMask,
                                     sigOrVal.updateTime)

    return HdlType.defaultConvert(self, sigOrVal, toType)



from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.struct import HStruct
from hwt.synthesizer.param import evalParam


class TransTmpl(object):
    """
    Container of informations about frames generated from any HType (HStruct etc.)

    * contains precalculated address range for this addrspace item
    """

    def __init__(self, dtype, bitAddr=0, parent=None, origin=None):
        """
        :ivar dtype: type of this item
        :ivar bitAddr: offset of start of this item in bits
        :ivar parent: object which generated this item
        :ivar origin: object which was template for generating of this item
        """
        self.parent = parent
        if origin is None:
            origin = dtype

        self.origin = origin
        self.dtype = dtype
        self.children = []
        self._loadFromHType(dtype, bitAddr)

    def _loadFromArray(self, dtype, bitAddr):
        self.itemCnt = evalParam(dtype.size).val
        self.children = TransTmpl(dtype.elmType, 0, self, origin=self.origin)
        return bitAddr + self.itemCnt * self.children.bitAddrEnd

    def _loadFromBits(self, dtype, bitAddr):
        return bitAddr + dtype.bit_length()

    def _loadFromHType(self, dtype, bitAddr):
        self.bitAddr = bitAddr

        if isinstance(dtype, HStruct):
            ld = self._loadFromHStruct
        elif isinstance(dtype, Array):
            ld = self._loadFromArray
        else:
            ld = self._loadFromBits

        self.bitAddrEnd = ld(dtype, bitAddr)

    def _loadFromHStruct(self, dtype, bitAddr):
        for f in dtype.fields:
            t = f.dtype
            origin = f
            isPadding = f.name is None

            if isPadding:
                width = t.bit_length()
                bitAddr += width
            else:
                fi = TransTmpl(t, bitAddr, parent=self, origin=origin)
                self.children.append(fi)
                bitAddr = fi.bitAddrEnd

        return bitAddr

    def getItemWidth(self):
        if not isinstance(self.dtype, Array):
            raise TypeError()
        return (self.bitAddrEnd - self.bitAddr) // self.itemCnt

    def bit_length(self):
        return self.bitAddrEnd - self.bitAddr

    def walkFlatten(self, offset=0,
                    shouldEnterFn=lambda transTmpl: (bool(transTmpl.children),
                                                     not bool(transTmpl.children))):
        """
        Walk fields in instance of TransTmpl

        :param offset: optional offset for all children in this TransTmpl
        :param shouldEnterFn: function (transTmpl) which returns True when field should
            be split on it's children
        :param shouldEnterFn: function(transTmpl) which should return (shouldEnter, shouldUse)
            where shouldEnter is flag that means iterator should look inside of this actual object
            and shouldUse flag means that this field should be used (=generator should yield it) 
        :return: generator of tuples ((startBitAddress, endBitAddress), TransTmpl instance)
        """

        t = self.dtype
        base = self.bitAddr + offset
        end = self.bitAddrEnd + offset

        shouldEnter, shouldYield = shouldEnterFn(self)
        if shouldYield:
            yield ((base, end), self)

        if shouldEnter:
            if isinstance(t, HStruct):
                for ch in self.children:
                    yield from ch.walkFlatten(offset=offset, shouldEnterFn=shouldEnterFn)
            elif isinstance(t, Array):
                itemSize = (self.bitAddrEnd - self.bitAddr) // self.itemCnt
                for i in range(self.itemCnt):
                    yield from self.children.walkFlatten(offset=base + i * itemSize,
                                                         shouldEnterFn=shouldEnterFn)
            elif isinstance(t, Bits):
                pass
            else:
                raise TypeError(t)

    def __repr__(self, offset=0):
        offsetStr = "".join(["    " for _ in range(offset)])

        try:
            name = self.origin.name
        except AttributeError:
            name = None

        if name:
            name = " name:%s," % name
        else:
            name = ""

        s = "%s<TransTmpl%s start:%d, end:%d" % (offsetStr, name, self.bitAddr, self.bitAddrEnd)
        if isinstance(self.dtype, Array):
            s += ", itemCnt:%d" % (self.itemCnt) + "\n"
            s += self.children.__repr__(offset=offset + 1) + "\n"
            s += offsetStr + ">"
            return s

        if not self.children:
            return s + ">"

        buff = [s, ]
        for child in self.children:
            buff.append(child.__repr__(offset=offset + 1))

        buff.append(offsetStr + ">")
        return "\n".join(buff)



from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.boolean import Boolean
from hwt.hdlObjects.types.defs import INT
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.hdlObjects.value import Value


def convertBits__val(self, sigOrVal, toType):
    if isinstance(toType, Boolean):
        return sigOrVal._eq(self.getValueCls().fromPy(1, self))
    elif isinstance(toType, Bits):
        if self.bit_length() == toType.bit_length():
            return sigOrVal._convSign(toType.signed)
    elif toType == INT:
        return INT.getValueCls()(sigOrVal.val,
                                 INT,
                                 int(sigOrVal._isFullVld()),
                                 sigOrVal.updateTime)

    return HdlType.defaultConvert(self, sigOrVal, toType)


def convertBits(self, sigOrVal, toType):
    if isinstance(sigOrVal, Value):
        return convertBits__val(self, sigOrVal, toType)
    if isinstance(toType, Boolean):
        if self.bit_length() == 1:
            v = 0 if sigOrVal._dtype.negated else 1
            return sigOrVal._eq(self.getValueCls().fromPy(v, self))
    elif isinstance(toType, Bits):
        if self.bit_length() == toType.bit_length():
            return sigOrVal._convSign(toType.signed)
    elif toType == INT:
        return Operator.withRes(AllOps.BitsToInt, [sigOrVal], toType)

    return HdlType.defaultConvert(self, sigOrVal, toType)





SYSTEMC_KEYWORDS = [
# c++11 keywords 
"alignas",
"alignof",
"and",
"and_eq",
"asm",
"atomic_cancel",
"atomic_commit",
"atomic_noexcept",
"auto",
"bitand",
"bitor",
"bool",
"break",
"case",
"catch",
"char",
"char16_t",
"char32_t",
"class",
"compl",
"concept",
"const",
"constexpr",
"const_cast",
"continue",
"decltype",
"default",
"delete",
"do",
"double",
"dynamic_cast",
"else",
"enum",
"explicit",
"export",
"extern",
"false",
"float",
"for",
"friend",
"goto",
"if",
"import",
"inline",
"int",
"long",
"module",
"mutable",
"namespace",
"new",
"noexcept",
"not",
"not_eq",
"nullptr",
"operator",
"or",
"or_eq",
"private",
"protected",
"public",
"register",
"reinterpret_cast",
"requires",
"return",
"short",
"signed",
"sizeof",
"static",
"static_assert",
"static_cast",
"struct",
"switch",
"synchronized",
"template",
"this",
"thread_local",
"throw",
"true",
"try",
"typedef",
"typeid",
"typename",
"union",
"unsigned",
"using",
"virtual",
"void",
"volatile",
"wchar_t",
"while",
"xor",
"xor_eq",
# SystemC namespace
"SC_MODULE", "SC_METHOD", "SC_THREAD", "SC_CTHREAD", "SC_CTOR", "SC_BIN", "SC_OCT", "SC_HEX", "SC_DEC",
"sensitive",
"sc_signal", "sc_in", "sc_out", "sc_inout", "sc_event", "sc_bit", "sc_fixed", "sc_fixed", "sc_lv", "sc_bv","sc_logic",
"sc_bigint", "sc_biguint",  "sc_int", "sc_out_resolved", "sc_inout_resolve", "sc_out_rv",
"sc_inout_rv", "sc_signal_resolved", "sc_signal_rv",
"notify", "sc_module", "sc_inout_clk", "sc_out_clk",
"sc_create_vcd_trace_file", "sc_trace_file", "sc_trace"]


from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdlObjects.operatorDefs import AllOps


# keep in mind that there is no such a thing in vhdl itself
opPrecedence = {AllOps.NOT: 2,
                AllOps.RISING_EDGE: 1,
                AllOps.NEG: 2,
                AllOps.DIV: 3,
                AllOps.ADD: 3,
                AllOps.SUB: 3,
                AllOps.MUL: 3,
                AllOps.XOR: 2,
                AllOps.EQ: 2,
                AllOps.NEQ: 2,
                AllOps.AND: 2,
                AllOps.OR: 2,
                AllOps.DOWNTO: 2,
                AllOps.GREATERTHAN: 2,
                AllOps.LOWERTHAN: 2,
                AllOps.GE: 2,
                AllOps.LE: 2,
                AllOps.CONCAT: 2,
                AllOps.INDEX: 1,
                AllOps.TERNARY: 1,
                AllOps.CALL: 1,
                }


def isResultOfTypeConversion(sig):
    try:
        sig.drivers[0]
    except IndexError:
        return False

    if sig.hidden:
        return True
    return False


class VhdlSerializer_ops():

    @classmethod
    def Operator(cls, op, ctx):
        def p(operand):
            s = cls.asHdl(operand, ctx)
            if isinstance(operand, RtlSignalBase):
                try:
                    o = operand.singleDriver()
                    if o.operator != op.operator and opPrecedence[o.operator] <= opPrecedence[op.operator]:
                        return "(%s)" % s
                except Exception:
                    pass
            return s
        # [TODO] no nested ternary in expressions like ( '1'  WHEN r = f ELSE  '0' ) & "0"
        ops = op.operands
        o = op.operator

        def _bin(name):
            return name.join(map(p, ops))

        if o == AllOps.AND:
            return _bin(' AND ')
        elif o == AllOps.OR:
            return _bin(' OR ')
        elif o == AllOps.XOR:
            return _bin(' XOR ')
        elif o == AllOps.NOT:
            assert len(ops) == 1
            return "NOT " + p(ops[0])
        elif o == AllOps.CALL:
            return "%s(%s)" % (cls.FunctionContainer(ops[0]), ", ".join(map(p, ops[1:])))
        elif o == AllOps.CONCAT:
            return _bin(' & ')
        elif o == AllOps.DIV:
            return _bin(' / ')
        elif o == AllOps.DOWNTO:
            return _bin('-1 DOWNTO ')
        elif o == AllOps.TO:
            return _bin('-1 TO ')
        elif o == AllOps.EQ:
            return _bin(' = ')
        elif o == AllOps.GREATERTHAN:
            return _bin(' > ')
        elif o == AllOps.GE:
            return _bin(' >= ')
        elif o == AllOps.LE:
            return _bin(' <= ')
        elif o == AllOps.INDEX:
            assert len(ops) == 2
            o1 = ops[0]
            if isinstance(o1, RtlSignalBase) and isResultOfTypeConversion(o1):
                o1 = ctx.createTmpVarFn("tmpTypeConv", o1._dtype)
                o1.defaultVal = ops[0]

            return "%s(%s)" % (cls.asHdl(o1, ctx).strip(), p(ops[1]))
        elif o == AllOps.LOWERTHAN:
            return _bin(' < ')
        elif o == AllOps.SUB:
            return _bin(' - ')
        elif o == AllOps.MUL:
            return _bin(' * ')
        elif o == AllOps.NEQ:
            return _bin(' /= ')
        elif o == AllOps.ADD:
            return _bin(' + ')
        elif o == AllOps.NEG:
            return "-(%s)" % (p(ops[0]))
        elif o == AllOps.TERNARY:
            return p(ops[1]) + " WHEN " + cls.condAsHdl([ops[0]], True, ctx) + " ELSE " + p(ops[2])
        elif o == AllOps.RISING_EDGE:
            assert len(ops) == 1
            return "RISING_EDGE(" + p(ops[0]) + ")"
        elif o == AllOps.FALLIGN_EDGE:
            assert len(ops) == 1
            return "FALLING_EDGE(" + p(ops[0]) + ")"
        elif o == AllOps.BitsAsSigned:
            assert len(ops) == 1
            return "SIGNED(" + p(ops[0]) + ")"
        elif o == AllOps.BitsAsUnsigned:
            assert len(ops) == 1
            return "UNSIGNED(" + p(ops[0]) + ")"
        elif o == AllOps.BitsAsVec:
            assert len(ops) == 1
            return "STD_LOGIC_VECTOR(" + p(ops[0]) + ")"
        elif o == AllOps.BitsToInt:
            assert len(ops) == 1
            op = cls.asHdl(ops[0], ctx)
            if ops[0]._dtype.signed is None:
                op = "UNSIGNED(%s)" % op
            return "TO_INTEGER(%s)" % op
        elif o == AllOps.IntToBits:
            assert len(ops) == 1
            resT = op.result._dtype
            op_str = cls.asHdl(ops[0], ctx)
            w = resT.bit_length()

            if resT.signed is None:
                return "STD_LOGIC_VECTOR(TO_UNSIGNED(" + op_str + ", %d))" % (w)
            elif resT.signed:
                return "TO_UNSIGNED(" + op_str + ", %d)" % (w)
            else:
                return "TO_UNSIGNED(" + op_str + ", %d)" % (w)

        elif o == AllOps.POW:
            assert len(ops) == 2
            return _bin(' ** ')
        else:
            raise NotImplementedError("Do not know how to convert %s to vhdl" % (o))



from hwt.hdlObjects.constants import DIRECTION
from hwt.hdlObjects.typeShortcuts import vecT
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import BIT, BIT_N
from hwt.interfaces.signalOps import SignalOps
from hwt.synthesizer.interfaceLevel.interface import Interface
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


D = DIRECTION


class Signal(SignalOps, Interface):
    """
    Basic wire interface
    """
    def __init__(self,
                 masterDir=D.OUT,
                 asArraySize=None,
                 dtype=BIT,
                 loadConfig=True):
        super().__init__(masterDir=masterDir,
                         asArraySize=asArraySize,
                         loadConfig=loadConfig)
        self._dtype = dtype

    def _injectMultiplerToDtype(self):
        """
        Make signal wider, used when there is an array of signals stored in one wider signal
        """
        t = self._dtype
        factor = self._widthMultiplier
        if t == BIT:
            newT = vecT(factor)
        elif isinstance(t, Bits):
            w = t.width
            if isinstance(w, int):
                newW = factor * w
            elif isinstance(w, RtlSignalBase):
                # both Param or factor Value
                newW = w * factor
            elif isinstance(factor, RtlSignalBase):
                # w is Value
                newW = factor * w
            else:
                # both Value
                newW = w.clone()
                newW.val *= factor.val
            newT = vecT(newW)
        else:
            raise TypeError("Can not multiply width of type %r" % (repr(t),))

        self._dtype = newT

    def _getSimAgent(self):
        from hwt.interfaces.agents.signal import SignalAgent
        return SignalAgent


def VectSignal(width,
               signed=None,
               masterDir=D.OUT,
               asArraySize=None,
               loadConfig=True):
    """
    Create basic :class:`.Signal` interface where type is vector
    """
    return Signal(masterDir,
                  asArraySize,
                  vecT(width, signed),
                  loadConfig)


class Clk(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as clock signal
    """
    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Clk
        return IP_Clk

    def _getSimAgent(self):
        from hwt.interfaces.agents.clk import OscilatorAgent
        return OscilatorAgent


class Rst(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as reset signal
    """
    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Rst
        return IP_Rst

    def _getSimAgent(self):
        from hwt.interfaces.agents.rst import PullDownAgent
        return PullDownAgent


class Rst_n(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as reset signal with negative polarity (active in 0)
    """
    def __init__(self,
                 masterDir=D.OUT,
                 asArraySize=None,
                 dtype=BIT_N,
                 loadConfig=True):
        super(Rst_n, self).__init__(masterDir=D.OUT,
                                    asArraySize=asArraySize,
                                    dtype=dtype,
                                    loadConfig=dtype)

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Rst_n
        return IP_Rst_n

    def _getSimAgent(self):
        from hwt.interfaces.agents.rst import PullUpAgent
        return PullUpAgent


class VldSynced(Interface):
    """
    Interface data+valid signal, if vld=1 then data are valid and slave should accept them
    """
    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = s(dtype=vecT(self.DATA_WIDTH))
        self.vld = s()

    def _getSimAgent(self):
        from hwt.interfaces.agents.vldSynced import VldSyncedAgent
        return VldSyncedAgent


class RdSynced(Interface):
    """
    Interface data+ready signal, if rd=1 then slave has read data and master should actualize data
    """
    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = s(dtype=vecT(self.DATA_WIDTH))
        self.rd = s(masterDir=D.IN)

    def _getSimAgent(self):
        from hwt.interfaces.agents.rdSynced import RdSyncedAgent
        return RdSyncedAgent


class Handshaked(VldSynced):
    """
    Interface data+ready+valid signal, if rd=1 slave is ready to accept data, if vld=1 master is sending data,
    if rd=1 and vld=1 then data is transfered otherwise master and slave has to wait on each other
    :attention: one rd/vld is set it must not go down until transaction is made
    """
    def _declr(self):
        super()._declr()
        self.rd = s(masterDir=D.IN)

    def _getSimAgent(self):
        from hwt.interfaces.agents.handshaked import HandshakedAgent
        return HandshakedAgent


class HandshakeSync(Interface):
    """
    Only synchronization interface, like vld+rd signal with meaning like in :class:`.Handshaked` interface

    :ivar rd: when high slave is ready to receive data
    :ivar vld: when high master is sending data to slave

    transaction happens when both ready and valid are high

    """
    def _declr(self):
        self.vld = s()
        self.rd = s(masterDir=D.IN)

    def _getSimAgent(self):
        from hwt.interfaces.agents.handshaked import HandshakeSyncAgent
        return HandshakeSyncAgent


class ReqDoneSync(Interface):
    """
    Synchronization interface, if req=1 slave begins operation and when its done it asserts done=1 for one clk tick
    req does not need to stay high
    """
    def _declr(self):
        self.req = s()
        self.done = s(masterDir=D.IN)


class BramPort_withoutClk(Interface):
    """
    Basic BRAM port
    """
    def _config(self):
        self.ADDR_WIDTH = Param(32)
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.addr = s(dtype=vecT(self.ADDR_WIDTH))
        self.din = s(dtype=vecT(self.DATA_WIDTH))
        self.dout = s(masterDir=D.IN, dtype=vecT(self.DATA_WIDTH))
        self.en = s()
        self.we = s()

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_BlockRamPort
        return IP_BlockRamPort

    def _getSimAgent(self):
        from hwt.interfaces.agents.bramPort import BramPort_withoutClkAgent
        return BramPort_withoutClkAgent

    def _getWordAddrStep(self):
        """
        :return: size of one word in unit of address
        """
        return 1

    def _getAddrStep(self):
        """
        :return: how many bits is one unit of address (f.e. 8 bits for  char * pointer,
             36 for 36 bit bram)
        """
        return int(self.DATA_WIDTH)


class BramPort(BramPort_withoutClk):
    """
    BRAM port with it's own clk
    """
    def _declr(self):
        self.clk = s(masterDir=D.OUT)
        with self._associated(clk=self.clk):
            super()._declr()

        self._associatedClk = self.clk

    def _getSimAgent(self):
        from hwt.interfaces.agents.bramPort import BramPortAgent
        return BramPortAgent


class FifoWriter(Interface):
    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.en = s()
        self.wait = s(masterDir=D.IN)
        self.data = s(dtype=vecT(self.DATA_WIDTH))

    def _getSimAgent(self):
        from hwt.interfaces.agents.fifo import FifoWriterAgent
        return FifoWriterAgent


class FifoReader(FifoWriter):
    def _declr(self):
        super()._declr()
        self.en._masterDir = D.IN
        self.wait._masterDir = D.OUT

    def _getSimAgent(self):
        from hwt.interfaces.agents.fifo import FifoReaderAgent
        return FifoReaderAgent


class RegCntrl(Interface):
    """
    Register control interface, :class:`.Signal` for read, :class:`.VldSynced` for write
    """
    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.din = Signal(dtype=vecT(self.DATA_WIDTH), masterDir=D.IN)
        with self._paramsShared():
            self.dout = VldSynced()

    def _getSimAgent(self):
        from hwt.interfaces.agents.regCntrl import RegCntrlAgent
        return RegCntrlAgent

s = Signal



from hwt.bitmask import mask
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import BOOL, BIT
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.hdlObjects.value import Value
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.value import GenericSerializer_Value
from hwt.serializer.serializerClases.constants import SIGNAL_TYPE
from hwt.serializer.serializerClases.indent import getIndent
from hwt.serializer.systemC.utils import systemCTypeOfSig
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class SystemCSerializer_value(GenericSerializer_Value):

    @classmethod
    def SignalItem(cls, si, ctx, declaration=False):
        sigType = systemCTypeOfSig(si)
        if declaration:
            if sigType is SIGNAL_TYPE.REG:
                fmt = "%s%s %s"
            else:
                fmt = "%ssc_signal<%s> %s"

            ctx = ctx.forSignal(si)

            v = si.defaultVal
            if si.virtualOnly:
                raise NotImplementedError()
            elif si.drivers:
                pass
            elif si.endpoints or si.simSensProcs:
                if not v.vldMask:
                    raise SerializerException("Signal %s is constant and has undefined value" % si.name)
            else:
                raise SerializerException("Signal %s should be declared but it is not used" % si.name)

            t = si._dtype
            dimensions = []
            while isinstance(t, Array):
                # collect array dimensions
                dimensions.append(t.size)
                t = t.elmType

            s = fmt % (getIndent(ctx.indent),
                       cls.HdlType(t, ctx),
                       si.name)
            if dimensions:
                # to make a space between name and dimensoins
                dimensions = list(map(lambda x: "[%s]" % cls.asHdl(toHVal(x), ctx),
                                      dimensions))
                dimensions.append("")
                s += " ".join(reversed(dimensions))

            if isinstance(v, RtlSignalBase):
                return s + " = %s" % cls.asHdl(v, ctx)
            elif isinstance(v, Value):
                if si.defaultVal.vldMask:
                    return s + " = %s" % cls.Value(si.defaultVal, ctx)
                else:
                    return s
            else:
                raise NotImplementedError(v)

        else:
            if si.hidden and hasattr(si, "origin"):
                return cls.asHdl(si.origin, ctx)
            else:
                if ctx.isTarget or sigType is SIGNAL_TYPE.REG:
                    return si.name
                else:
                    return "%s.read()" % si.name

    @classmethod
    def condAsHdl(cls, cond, forceBool, ctx):
        if isinstance(cond, RtlSignalBase):
            cond = [cond]
        else:
            cond = list(cond)
        if len(cond) == 1:
            c = cond[0]
            if not forceBool or c._dtype == BOOL:
                return cls.asHdl(c, ctx)
            elif c._dtype == BIT:
                return cls.asHdl(c._eq(1), ctx)
            elif isinstance(c._dtype, Bits):
                return cls.asHdl(c != 0, ctx)
            else:
                raise NotImplementedError()
        else:
            return " && ".join(map(lambda x: cls.condAsHdl(x, forceBool, ctx), cond))

    @classmethod
    def BitString(cls, v, width, vldMask=None):
        if vldMask is None:
            vldMask = mask(width)
        # if can be in hex
        if width % 4 == 0 and vldMask == (1 << width) - 1:
            t = cls.HdlType_bits(Bits(width), None)
            return ('%s("0x%0' + str(width // 4) + 'x")') % (t, v)
        else:  # else in binary
            return cls.BitString_binary(v, width, vldMask)

    @classmethod
    def BitLiteral(cls, v, vldMask):
        if vldMask:
            return "'%d'" % int(bool(v))
        else:
            return "'X'"

    @classmethod
    def BitString_binary(cls, v, width, vldMask=None):
        t = cls.HdlType_bits(Bits(width), None)
        buff = [t, '("']
        for i in range(width - 1, -1, -1):
            mask = (1 << i)
            b = v & mask

            if vldMask & mask:
                s = "1" if b else "0"
            else:
                s = "X"
            buff.append(s)
        buff.append('")')
        return ''.join(buff)

    @classmethod
    def _BitString(cls, typeName, v, width, forceVector, vldMask):
        if vldMask != mask(width):
            if forceVector or width > 1:
                v = cls.BitString(v, width, vldMask)
            else:
                v = cls.BitLiteral(v, width, vldMask)
        else:
            v = str(v)
        # [TODO] parametrized width
        return "%s<%d>(%s)" % (typeName, width, v)

    @classmethod
    def Enum_valAsHdl(cls, dtype, val, ctx):
        i = dtype._allValues.index(val.val)
        assert i >= 0
        return '%d' % i

    @classmethod
    def SignedBitString(cls, v, width, forceVector, vldMask):
        return cls._BitString("sc_biguint", v, width, forceVector, vldMask)

    @classmethod
    def UnsignedBitString(cls, v, width, forceVector, vldMask):
        return cls._BitString("sc_biguint", v, width, forceVector, vldMask)

    @classmethod
    def Array_valAsHdl(cls, dtype, val, ctx):
        separator = ",\n" + getIndent(ctx.indent + 1)
        return "".join(["{", separator.join([cls.Value(v, ctx) for v in val]), "}"])




class TransactionPart(object):
    """
    Container for informations about parts of TransTmpl split on databus words
    """
    def __init__(self, tmpl, startOfPart, endOfPart, inFieldOffset):
        self.tmpl = tmpl
        self.isPadding = tmpl is None
        self.startOfPart = startOfPart
        self.endOfPart = endOfPart
        self.inFieldOffset = inFieldOffset
        self.parent = None

    def bit_length(self):
        return self.endOfPart - self.startOfPart

    def getBusWordBitRange(self):
        """
        :return: bit range which contains data of this part on bus data signal
        """
        offset = self.startOfPart % self.parent.wordWidth
        return (offset + self.bit_length(), offset)

    def getFieldBitRange(self):
        """
        :return: bit range which contains data of this part on interface of field
        """
        offset = self.inFieldOffset
        return (self.bit_length() + offset, offset)

    def isLastPart(self):
        return self.tmpl.bitAddrEnd == self.endOfPart

    def __repr__(self):
        return "<TransactionPart %r, startOfPart:%d, endOfPart:%d>" % (
               self.tmpl, self.startOfPart, self.endOfPart)



from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.value import Value
from hwt.hdlObjects.statements import IfContainer, SwitchContainer
from hwt.pyUtils.arrayQuery import areSetsIntersets, groupedby
from hwt.serializer.utils import maxStmId
from itertools import islice, zip_longest


def removeUnconnectedSignals(netlist):
    """
    If signal is not driving anything remove it
    """

    toDelete = set()
    toSearch = netlist.signals

    while toSearch:
        _toSearch = set()

        for sig in toSearch:
            if not sig.endpoints:
                try:
                    if sig._interface is not None:
                        # skip interfaces before we want to check them,
                        # they should not be optimized out from design
                        continue
                except AttributeError:
                    pass

                for e in sig.drivers:
                    # drivers of this signal are useless rm them
                    if isinstance(e, Operator):
                        for op in e.operands:
                            if not isinstance(op, Value):
                                op.endpoints.remove(e)
                                _toSearch.add(op)

                    elif isinstance(e, Assignment):
                        op = e.src
                        if not isinstance(op, Value):
                            op.endpoints.remove(e)
                            _toSearch.add(op)

                        netlist.startsOfDataPaths.remove(e)
                    else:
                        assert False, ("Drivers should be only index operators or assignments", e)

                toDelete.add(sig)

        if toDelete:
            for sig in toDelete:
                netlist.signals.remove(sig)
                try:
                    _toSearch.remove(sig)
                except KeyError:
                    pass
            toDelete = set()
        toSearch = _toSearch


class IncompatibleStructure(Exception):
    """
    instances of HWProcess can not be merged due incompatible structure
    """


def isMergableStmList(listA, listB):
    la = len(listA)
    lb = len(listB)
    if la == lb:
        return True
    elif la > 0 and lb == 1:
        return True
    else:
        return False


def checkIfIsTooSimple(proc):
    """check if process is just unconditional assignments and it is useless to merge them"""
    try:
        a, = proc.statements
        if isinstance(a, Assignment):
            return True
    except ValueError:
        pass
    return False


def mergeStmLists(stmsA, stmsB):
    tmp = []
    for a, b in zip_longest(stmsA, stmsB, fillvalue=None):
        if b is None:
            assert a is not None
            tmp.append(a)
        else:
            tmp.extend(tryToMergeStm(a, b))

    return tmp


def tryToMergeStm(stmA, stmB):
    """
    :raise IncompatibleStructure: if it is not possible to merge statements
    :return: generator of statements
    """

    if isinstance(stmA, Assignment) or isinstance(stmB, Assignment):
        yield stmA
        yield stmB
        return

    aIsIf = isinstance(stmA, IfContainer)
    bIsIf = isinstance(stmB, IfContainer)
    if aIsIf and bIsIf:
        if (stmA.cond != stmB.cond or
                not isMergableStmList(stmA.ifTrue, stmB.ifTrue) or
                not isMergableStmList(stmA.elIfs, stmB.elIfs) or
                not isMergableStmList(stmA.ifFalse, stmB.ifFalse)):
            raise IncompatibleStructure()

        ifTrue = mergeStmLists(stmA.ifTrue, stmB.ifTrue)

        elIfs = []
        for (condA, elifA), (condB, elifB) in zip(stmA.elIfs, stmB.elIfs):
            if (condA != condB or not isMergableStmList(elifA, elifB)):
                raise IncompatibleStructure()
            elIfs.append((condA, mergeStmLists(elifA, elifB)))

        ifFalse = mergeStmLists(stmA.ifFalse, stmB.ifFalse)

        yield IfContainer(stmA.cond, ifTrue, ifFalse, elIfs)
        return

    aIsSwitch = isinstance(stmA, SwitchContainer)
    bIsSwitch = isinstance(stmB, SwitchContainer)
    if (aIsSwitch and bIsSwitch and
            stmA.switchOn is stmB.switchOn and
            len(stmA.cases) == len(stmB.cases) and
            isMergableStmList(stmA.default, stmB.default)):

        cases = []
        for (vA, caseA), (vB, caseB) in zip(stmA.cases, stmB.cases):
            if vA != vB or not isMergableStmList(caseA, caseB):
                raise IncompatibleStructure()
            cases.append((vA, mergeStmLists(caseA, caseB)))

        default = mergeStmLists(stmA.default, stmB.default)

        yield SwitchContainer(stmA.switchOn, cases, default)
        return

    raise NotImplementedError(stmA, stmB)


def tryToMerge(procA, procB):
    """
    Try merge procB into procA

    :raise IncompatibleStructure: if merge is not possible
    :attention: procA is now result if merge has succeed
    :return: procA which is now result of merge
    """
    if (checkIfIsTooSimple(procA) or
            checkIfIsTooSimple(procB) or
            areSetsIntersets(procA.outputs, procB.sensitivityList) or
            areSetsIntersets(procB.outputs, procA.sensitivityList) or
            len(procA.statements) != len(procB.statements)):
        raise IncompatibleStructure()

    statements = []
    for stmA, stmB in zip(procA.statements, procB.statements):
        statements.extend(tryToMergeStm(stmA, stmB))

    procA.statements = statements
    procA.outputs.update(procB.outputs)
    procA.inputs.update(procB.inputs)
    procA.sensitivityList.update(procB.sensitivityList)

    return procA


def reduceProcesses(processes, procRanks):
    """
    Try to merge processes as much is possible

    :param processes: list of processes instances
    :param procRanks: process ranks = how many assignments is probably in process
        used to minimize number of merge tries
    """
    # sort to make order of merging same deterministic
    processes.sort(key=lambda x: (x.name, maxStmId(x)))
    # now try to reduce processes with nearly same structure of statements into one
    # to minimize number of processes
    for _, procs in groupedby(processes, lambda p: procRanks[p]):
        for iA, pA in enumerate(procs):
            if pA is None:
                continue
            for iB, pB in enumerate(islice(procs, iA + 1, None)):
                if pB is None:
                    continue
                try:
                    pA = tryToMerge(pA, pB)
                except IncompatibleStructure:
                    continue
                procs[iA + 1 + iB] = None
                # procs[iA] = pA

        for p in procs:
            if p is not None:
                yield p



from hwt.interfaces.std import Rst_n
from hwt.simulator.shortcuts import onRisingEdge
from hwt.synthesizer.exceptions import IntfLvlConfErr


class AgentBase():
    """
    Base class of agent of interface like in UVM
    driver is used for slave interfaces
    monitor is used for master interfaces

    :ivar intf: interface assigned to this agent
    :ivar enable: flag to enable/disable this agent
    """
    def __init__(self, intf):
        self.intf = intf
        self.enable = True
        self._debugOutput = None

    def _debug(self, out):
        self._debugOutput = out

    def getDrivers(self):
        """
        Called before simulation to collect all drivers of interfaces from this agent
        """
        return [self.driver]

    def getMonitors(self):
        """
        Called before simulation to collect all monitors of interfaces from this agent
        """
        return [self.monitor]

    def driver(self, s):
        """
        Implement this method to drive your interface in simulation/verification
        """
        raise NotImplementedError()

    def monitor(self, s):
        """
        Implement this method to monitor your interface in simulation/verification
        """
        raise NotImplementedError()


class SyncAgentBase(AgentBase):
    """
    Agent which discovers clk, rst signal and runs only at specified edge of clk

    :attention: requires clk and rst/rstn signal
        (if you do not have any create simulation wrapper with it)
    """
    def __init__(self, intf, allowNoReset=False):
        super().__init__(intf)

        # resolve clk and rstn
        self.clk = self.intf._getAssociatedClk()._sigInside
        try:
            self.rst = self.intf._getAssociatedRst()
            self.rstOffIn = isinstance(self.rst, Rst_n)
            self.rst = self.rst._sigInside
        except IntfLvlConfErr as e:
            self.rst = None
            if allowNoReset:
                pass
            else:
                raise e

        # run monitor, driver only on rising edge of clk
        self.monitor = onRisingEdge(self.clk, self.monitor)
        self.driver = onRisingEdge(self.clk, self.driver)

    def notReset(self, s):
        rst = self.rst
        if rst is None:
            return True
        else:
            rstVal = s.read(rst).val
            return bool(rstVal) == self.rstOffIn



from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.types.sliceUtils import slice_to_SLICE
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.hdlObjects.value import Value
from hwt.synthesizer.exceptions import TypeConversionErr
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc


def tv(signal):
    """
    Value class for type of signal
    """
    return signal._dtype.getValueCls()


class RtlSignalOps():
    """
    Definitions of operators and other operator functions for RtlSignal

    :ivar _usedOps: cache for expressions with this signal
    """

    def _convert(self, toT):
        return tv(self)._convert(self, toT)

    def naryOp(self, operator, opCreateDelegate, *otherOps):
        """
        Try lookup operator with this parameters in _usedOps
        if not found create new one and soter it in _usedOps
        """
        k = (operator, *otherOps)
        try:
            return self._usedOps[k]
        except KeyError:
            o = opCreateDelegate(self, *otherOps)
            self._usedOps[k] = o
            return o

    def __invert__(self):
        return self.naryOp(AllOps.NOT, tv(self).__invert__)

    def _onRisingEdge(self, now=None):
        return self.naryOp(AllOps.RISING_EDGE, tv(self)._onRisingEdge, now)

    def _onFallingEdge(self, now=None):
        return self.naryOp(AllOps.FALLIGN_EDGE, tv(self)._onFallingEdge, now)

    def _isOn(self):
        return self._dtype.convert(self, BOOL)

    # conversions
    def _convSign(self, signed):
        return tv(self)._convSign(self, signed)

    def _signed(self):
        return tv(self)._signed(self)

    def _unsigned(self):
        return tv(self)._unsigned(self)

    def _vec(self):
        return tv(self)._vec(self)

    # logic
    def __and__(self, other):
        return self.naryOp(AllOps.AND, tv(self).__and__, other)

    def __xor__(self, other):
        return self.naryOp(AllOps.XOR, tv(self).__xor__, other)

    def __or__(self, other):
        return self.naryOp(AllOps.OR, tv(self).__or__, other)

    # cmp
    def _eq(self, other):
        """__eq__ is not overloaded because it will destroy hashability of object"""
        return self.naryOp(AllOps.EQ, tv(self)._eq, other)

    def __ne__(self, other):
        return self.naryOp(AllOps.NEQ, tv(self).__ne__, other)

    def __ge__(self, other):
        return self.naryOp(AllOps.GE, tv(self).__ge__, other)

    def __gt__(self, other):
        return self.naryOp(AllOps.GREATERTHAN, tv(self).__gt__, other)

    def __lt__(self, other):
        return self.naryOp(AllOps.LOWERTHAN, tv(self).__lt__, other)

    def __le__(self, other):
        return self.naryOp(AllOps.LE, tv(self).__le__, other)

    # arithmetic
    def __add__(self, other):
        return self.naryOp(AllOps.ADD, tv(self).__add__, other)

    def __sub__(self, other):
        return self.naryOp(AllOps.SUB, tv(self).__sub__, other)

    def __mul__(self, other):
        return self.naryOp(AllOps.MUL, tv(self).__mul__, other)

    def __floordiv__(self, divider):
        return self.naryOp(AllOps.DIV, tv(self).__floordiv__, divider)

    # selections
    def _downto(self, to):
        return self.naryOp(AllOps.DOWNTO, tv(self)._downto, to)

    def __getitem__(self, key):
        operator = AllOps.INDEX
        if isinstance(key, slice):
            key = slice_to_SLICE(key, self._dtype.bit_length())

        k = (operator, key)
        try:
            return self._usedOps[k]
        except KeyError:
            o = tv(self).__getitem__(self, key)
            self._usedOps[k] = o
            return o

    def _concat(self, *operands):
        return self.naryOp(AllOps.CONCAT, tv(self)._concat, *operands)

    def _ternary(self, ifTrue, ifFalse):
        return self.naryOp(AllOps.TERNARY, tv(self)._ternary, ifTrue, ifFalse)

    def _getIndexCascade(self):
        """
        Find out if this signal is something indexed
        """
        try:
            # now I am result of the index  xxx[xx] <= source
            # get index op
            d = self.singleDriver()
            try:
                op = d.operator
            except AttributeError:
                op = None
            if op == AllOps.INDEX:
                # get signal on which is index applied
                indexedOn = d.operands[0]
                if isinstance(indexedOn, RtlSignalBase):
                    # [TODO] multidimensional indexing
                    return indexedOn, [d.operands[1]]
                else:
                    raise Exception("can not drive static value %r" % indexedOn)

        except MultipleDriversExc:
            pass

    def __pow__(self, source):
        """
        Create assignment to this signal

        :attention: it is not power operator it is assignment
        :return: list of assignments
        """
        if isinstance(source, InterfaceBase):
            assert source._isAccessible
            source = source._sig

        if source is None:
            source = self._dtype.fromPy(None)
        else:
            source = toHVal(source)
            err = False
            try:
                source = source._dtype.convert(source, self._dtype)
            except TypeConversionErr:
                err = True
            if err:
                raise TypeConversionErr(("Can not connect %r (of type %r) to %r "
                                        "(of type %r) due type incompatibility")
                                        % (source, source._dtype, self, self._dtype))

        tmp = self._getIndexCascade()
        if tmp:
            mainSig, indexCascade = tmp
            self = mainSig
        else:
            indexCascade = None

        # self = self._tryMyIndexToEndpoint()
        a = Assignment(source, self, indexCascade)

        self.drivers.append(a)
        if not isinstance(source, Value):
            source.endpoints.append(a)

        return [a]

    def __int__(self):
        if not self._const:
            raise TypeError("Int value of signal can be evaluated"
                            " because it is not constant expression:", self)
        else:
            return int(self._val)

    def __bool__(self):
        if not self._const:
            raise TypeError("Bool value of signal can be evaluated"
                            "because it is not constant expression:", self)
        else:
            return bool(self._val)

    def _isFullVld(self):
        return self._const and self._val._isFullVld()



from hwt.hdlObjects.types.hdlType import HdlType


class Boolean(HdlType):

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return hash(Boolean)

    @classmethod
    def getConvertor(cls):
        from hwt.hdlObjects.types.booleanConversions import convertBoolean
        return convertBoolean

    def bit_length(self):
        return 1

    def all_mask(self):
        return 1

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.booleanVal import BooleanVal
            cls._valCls = BooleanVal
            return cls._valCls



VERILOG_KEYWORDS = [
                    "always"
                    "and",
                    "assign",
                    "begin",
                    "buf",
                    "bufif0",
                    "bufif1",
                    "case",
                    "casex",
                    "casez",
                    "cmos",
                    "deassign",
                    "default",
                    "defparam",
                    "disable",
                    "edge",
                    "else",
                    "end",
                    "endcase",
                    "endfunction",
                    "endmodule",
                    "endprimitive",
                    "endspecify",
                    "endtable",
                    "endtask",
                    "event",
                    "for",
                    "force",
                    "forever",
                    "fork",
                    "function",
                    "highz0",
                    "highz1",
                    "if",
                    "initial",
                    "inout",
                    "input",
                    "integer",
                    "join",
                    "large",
                    "macromodule",
                    "medium",
                    "module",
                    "nand",
                    "negedge",
                    "nmos",
                    "nor",
                    "not",
                    "notif0",
                    "notif1",
                    "or",
                    "output",
                    "pmos",
                    "posedge",
                    "primitive",
                    "pull0",
                    "pull1",
                    "pulldown",
                    "pullup",
                    "rcmos",
                    "reg",
                    "release",
                    "repeat",
                    "rnmos",
                    "rpmos",
                    "rtran",
                    "rtranif0",
                    "rtranif1",
                    "scalared",
                    "small",
                    "specify",
                    "specparam",
                    "strong0",
                    "strong1",
                    "supply0",
                    "supply1",
                    "table",
                    "task",
                    "time",
                    "tran",
                    "tranif0",
                    "tranif1",
                    "tri",
                    "tri0",
                    "tri1",
                    "triand",
                    "trior",
                    "vectored",
                    "wait",
                    "wand",
                    "weak0",
                    "weak1",
                    "while",
                    "wire",
                    "wor",
                    "xnor",
                    "xor",
                    ]


from hwt.hdlObjects.types.hdlType import HdlType
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class Array(HdlType):
    """
    vldMask and eventMask on Array_val instance is not used instead of that
    these flags on elements are used
    """
    def __init__(self, elmType, size):
        super(Array, self).__init__()
        self.elmType = elmType
        self.size = size

    def __eq__(self, other):
        return (
            type(self) is type(other) and
            self.elmType == other.elmType and
            self.size == other.size
            )

    def __hash__(self):
        return hash((self.elmType, self.size))

    def bit_length(self):
        """
        :return: bit width for this type
        """
        try:
            itemSize = self.elmType.bit_length
        except AttributeError:
            itemSize = None
        if itemSize is None:
            raise TypeError("Can not determine size of array because item has not determinable size")

        s = self.size
        if isinstance(s, RtlSignalBase):
            s = int(s.staticEval())
        return s * itemSize()

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.arrayVal import ArrayVal
            cls._valCls = ArrayVal
            return cls._valCls

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not none is used as a additional information about where
            on which address this type is stored (used only by HStruct)
        :param expandStructs: expand HStruct types (used by HStruct and Array)
        """

        return "%s[%r]" % (self.elmType.__repr__(indent=indent,
                                                 withAddr=withAddr,
                                                 expandStructs=expandStructs),
                           self.size)



from hwt.synthesizer.exceptions import TypeConversionErr


class HdlType():
    """
    Base class for all hardware related types.
    """

    def fromPy(self, v):
        """
        Construct value of this type.
        Delegated on value class for this type
        """
        return self.getValueCls().fromPy(v, self)

    def convert(self, sigOrVal, toType):
        """
        Cast value or signal of this type to another.
        :param sigOrVal: instance of signal or value to cast
        :param toType: instance of HdlType to cast into
        """
        if sigOrVal._dtype == toType:
            return sigOrVal

        try:
            c = self._convert
        except AttributeError:
            c = self.getConvertor()
            self._convertor = c

        return c(self, sigOrVal, toType)

    @classmethod
    def getConvertor(cls):
        """
        Get method for converting type
        """
        return HdlType.defaultConvert

    def defaultConvert(self, sigOrVal, toType):
        raise TypeConversionErr("Conversion of %r of type \n%r to type %r is not implemented"
                                % (sigOrVal, self, toType))

    @classmethod
    def getValueCls(cls):
        """
        :attention: Overrode in implementation of concrete HdlType.

        :return: class for value derived from this type
        """
        raise NotImplementedError()

    def __getitem__(self, key):
        """
        [] operator to create an array of this type.
        """
        assert int(key) > 0, key # array has to have some items
        from hwt.hdlObjects.types.array import Array
        return Array(self, key)

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not none is used as a additional information about where
            on which address this type is stored (used only by HStruct)
        :param expandStructs: expand HStruct types (used by HStruct and Array)
        """
        return "<%s>" % (self.__class__.__name__)



from hwt.interfaces.std import Clk, Rst_n, Rst
from hwt.code import connect


def addClkRstn(self):
    self.clk = Clk()
    self.rst_n = Rst_n()


def addClkRst(self):
    self.clk = Clk()
    self.rst = Rst()


def _tryConnect(src, unit, intfName):
    "try connect src to interface of specified name on unit"
    try:
        dst = getattr(unit, intfName)
    except AttributeError:
            dst = None
    if dst is not None:
        connect(src, dst)


def propagateClk(self):
    """
    Propagate "clk" clock signal to all subcomponents
    """
    clk = self.clk
    for u in self._units:
        _tryConnect(clk, u, 'clk')


def propagateClkRstn(self):
    """
    Propagate "clk" clock and negative reset "rst_n" signal to all subcomponents
    """
    clk = self.clk
    rst_n = self.rst_n

    for u in self._units:
        _tryConnect(clk, u, 'clk')
        _tryConnect(rst_n, u, 'rst_n')
        _tryConnect(~rst_n, u, 'rst')


def propagateClkRst(self):
    """
    Propagate "clk" clock and reset "rst" signal to all subcomponents
    """
    clk = self.clk
    rst = self.rst

    for u in self._units:
        _tryConnect(clk, u, 'clk')
        _tryConnect(~rst, u, 'rst_n')
        _tryConnect(rst, u, 'rst')


import math

from hwt.hdlObjects.typeShortcuts import vec
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.struct import HStruct
from hwt.hdlObjects.types.structUtils import walkFlattenFields


def fitTo(what, to):
    """
    Slice signal "what" to fit in "to"
    or
    extend "what" with zeros to same width as "to"

    little-endian impl.
    """

    whatWidth = what._dtype.bit_length()
    toWidth = to._dtype.bit_length()
    if toWidth == whatWidth:
        return what
    elif toWidth < whatWidth:
        # slice
        return what[toWidth:]
    else:
        if what._dtype.signed:
            raise NotImplementedError("Signed extension")
        # extend
        return vec(0, toWidth - whatWidth)._concat(what)


def iterBits(sigOrVal, bitsInOne=1, skipPadding=True, fillup=False):
    """
    Iterate over bits in vector

    :param sig: signal or value to iterate over
    :param bitsInOne: number of bits in one part
    :param skipPadding: if true padding is skipped in dense types
    """
    if isinstance(bitsInOne, int):
        bitsInOne = int(bitsInOne)

    t = sigOrVal._dtype
    if isinstance(t, (HStruct, Array)):
        actual = None
        actualOffset = 0

        for f in walkFlattenFields(sigOrVal, skipPadding=skipPadding):
            thisFieldLen = f._dtype.bit_length()

            if actual is None:
                actual = f
                actuallyHave = thisFieldLen
            else:
                bitsInActual = actual._dtype.bit_length() - actualOffset
                actuallyHave = bitsInActual + thisFieldLen
                if actuallyHave >= bitsInOne:
                    # consume what was remained in actual
                    takeFromThis = bitsInOne - bitsInActual
                    yield f[takeFromThis:]._concat(actual[:actualOffset])
                    actualOffset = takeFromThis
                    actuallyHave -= bitsInOne
                    if actuallyHave > 0:
                        actual = f
                    else:
                        actual = None
                else:
                    # concat to actual because it is not enough
                    actual = f._concat(actual)

            while actuallyHave >= bitsInOne:
                yield actual[(actualOffset + bitsInOne):actualOffset]
                # update slice out what was taken
                actuallyHave -= bitsInOne
                actualOffset += bitsInOne

            if actuallyHave == 0:
                actual = None
                actualOffset = 0

        if actual is not None and fillup:
            fillupW = bitsInOne - actuallyHave
            t = f._dtype
            padding = Bits(fillupW, signed=t.signed, negated=t.negated).fromPy(None)
            yield padding._concat(actual[:actualOffset])
        else:
            assert actual is None, "Width of object has to be divisible by bitsInOne"
    else:
        w = sigOrVal._dtype.bit_length()
        if bitsInOne == 1:
            for bit in range(w):
                yield sigOrVal[bit]
        else:
            items = math.ceil(w / bitsInOne)
            for i in range(items):
                h = min(bitsInOne * (i + 1), w)
                l = bitsInOne * i
                s = sigOrVal[h:l]
                if i < items - 1 or s._dtype.bit_length() == bitsInOne:
                    yield s
                else:
                    # is last and is not complete
                    assert fillup
                    _w = s._dtype.bit_length()
                    padding = Bits(bitsInOne - _w).fromPy(None)
                    yield padding._concat(s)



from collections import deque

from hwt.hdlObjects.constants import NOP
from hwt.simulator.agentBase import SyncAgentBase


class VldSyncedAgent(SyncAgentBase):
    def __init__(self, intf, allowNoReset=False):
        super(VldSyncedAgent, self).__init__(intf, allowNoReset=allowNoReset)
        self.data = deque()

    def doRead(self, s):
        return s.read(self.intf.data)

    def doWrite(self, s, data):
        s.write(data, self.intf.data)

    def monitor(self, s):
        intf = self.intf
        yield s.updateComplete
        if self.enable and self.notReset(s):
            vld = s.read(intf.vld)
            assert vld.vldMask, "valid signal for interface %r is in invalid state, this would cause desynchronization in %d" % (self.intf, s.now)
            if vld.val:
                d = self.doRead(s)

                if self._debugOutput is not None:
                    self._debugOutput.write("%s, read, %d: %r\n" % (
                                              self.intf._getFullName(), s.now, d))
                self.data.append(d)

    def driver(self, s):
        intf = self.intf

        if self.enable and self.data and self.notReset(s):
            d = self.data.popleft()
            if d is NOP:
                self.doWrite(s, None)
                s.write(0, intf.vld)
            else:
                self.doWrite(s, d)
                s.write(1, intf.vld)
                if self._debugOutput is not None:
                    self._debugOutput.write("%s, wrote, %d: %r\n" % (
                                                self.intf._getFullName(), s.now, self.actualData))

        else:
            self.doWrite(s, None)
            s.write(0, intf.vld)



from simpy.core import BoundClass
from simpy.events import NORMAL, Timeout

from hwt.hdlObjects.value import Value
from hwt.simulator.hdlSimConfig import HdlSimConfig
from hwt.simulator.simModel import mkUpdater, mkArrayUpdater
from hwt.simulator.simulatorCore import HdlEnvironmentCore
from hwt.simulator.utils import valueHasChanged
from hwt.synthesizer.uniqList import UniqList


def isEvDependentOn(sig, process):
    if sig is None:
        return False
    return process in sig.simFallingSensProcs or process in sig.simRisingSensProcs


class UpdateSet(set):
    """
    Set of updates for signal

    :ivar destination: signal which are updates for
    """
    def __init__(self, destination):
        self.destination = destination


class IoContainer():
    """
    Container for outputs of process
    """
    __slots__ = ["_all_signals"]

    def __init__(self, dstSignalsTuples):
        """
        :param dstSignalsTuples: tuples (name, signal)
        """
        self._all_signals = []
        for name, s in dstSignalsTuples:
            o = UpdateSet(s)
            setattr(self, name, o)
            self._all_signals.append(o)


class HdlSimulator(HdlEnvironmentCore):
    """
    Circuit simulator with support for external agents

    .. note::
        Every signal is initialized at start with its default value
        (sig. without driver, sig with constant driver solved)

    .. note::
        Every interprocess signal is marked by synthesizer and it can not be directly updated
        by any process, process should only return tuple (updateDestionation, updateFn, isEventDependentFlag)
        and let simulator to update it for others, any other signals are evaluated as expression
        by every process
        every process drives only one signal
        every process uses sensitivity-list like in other languages (but it is generated automatically)
        (communication between process solved)

    .. note::
        Hdlprocesses can not contain any wait statements etc. only simulation processes can.
        Simulation processes are written in python.
        (using hdl as main simulator driver is not efficient and thats why it is not supported
        and it is easy to just read hdl process with unsupported statements and translate them to
        simulator commands)

    .. note::
        HWprocesses have lower priority than simulation processes this allows simplify logic of all agents
        when simulation process is executed HW part did not anything in this time
        so simulation process can prepare anything for HW part (= can write)
        if simulation process need to read, it has to yield simulator.updateComplete
        first, process then will be waken after reaction of HW in this time:
        agents are greatly simplified, they just need to yield simulator.updateComplete
        before first read and then the can not write in this time

    :ivar now: actual simulation time
    :ivar updateComplete: this event is triggered when there are not any values to apply in this time
    :ivar valuesToApply: is container to for quantum of values which should be applied in single time
    :ivar env: simply environment
    :ivar applyValuesPlaned: flag if there is planed applyValues for current values quantum
    :ivar seqProcsToRun: list of event dependent processes which should be evaluated after applyValEv
    """
    PRIORITY_APPLY_COMB = NORMAL + 1
    PRIORITY_APPLY_SEQ = PRIORITY_APPLY_COMB + 1

    def __init__(self, config=None):
        super(HdlSimulator, self).__init__()
        if config is None:
            # default config
            config = HdlSimConfig()

        self.config = config
        self.updateComplete = self.event()
        self.applyValEv = None
        self.runSeqProcessesEv = None

        # (signal, value) tuples which should be applied before
        # new round of processes
        #  will be executed
        self.valuesToApply = []
        self.seqProcsToRun = UniqList()
        self.combProcsToRun = UniqList()
        # container of outputs for every process
        self.outputContainers = {}

    def addHwProcToRun(self, trigger, proc):
        # first process in time has to plan executing of apply values on the end of this time
        if self.applyValEv is None:
            # (apply on end of this time to minimalize process reevaluation)
            self.scheduleApplyValues()

        if isEvDependentOn(trigger, proc):
            if self.now == 0:
                return  # pass event dependent on startup
            self.seqProcsToRun.append(proc)
        else:
            self.combProcsToRun.append(proc)

    def _initUnitSignals(self, unit):
        """
        * Inject default values to simulation

        * Instantiate IOs for every process
        """
        for s in unit._cntx.signals:
            v = s.defaultVal.clone()

            # force update all signals to deafut values and propagate it
            s.simUpdateVal(self, mkUpdater(v, False))

        for u in unit._units:
            self._initUnitSignals(u)

        for p in unit._processes:
            self.addHwProcToRun(None, p)

        for p, outputs in unit._outputs.items():
            # name has to be explicit because it may be possible that signal
            # with has this name was replaced by signal from parent/child
            containerNames = list(map(lambda x: x[0], outputs))

            class SpecificIoContainer(IoContainer):
                __slots__ = containerNames

            self.outputContainers[p] = SpecificIoContainer(outputs)

    def scheduleApplyValues(self):
        applyVal = self.applyValEv = self.event()
        applyVal._ok = True
        applyVal._value = None
        applyVal.callbacks.append(self.applyValues)

        self.schedule(applyVal, priority=self.PRIORITY_APPLY_COMB)

        if self.runSeqProcessesEv is not None:
            return

        assert not self.seqProcsToRun
        runSeq = self.runSeqProcessesEv = self.event()
        runSeq._ok = True
        runSeq._value = None
        runSeq.callbacks.append(self.runSeqProcesses)

        self.schedule(runSeq, priority=self.PRIORITY_APPLY_SEQ)

    def conflictResolveStrategy(self, actionSet):
        """
        This functions resolves

        :param actionSet: set of actions made by process
        """
        invalidate = False
        l = len(actionSet)
        # resolve if there is no write collision
        if l == 0:
            return
        elif l == 1:
            res = actionSet.pop()
        else:
            # we are driving signal with two different values so we invalidate result
            res = actionSet.pop()
            invalidate = True

        l = len(res)
        if l == 3:
            # update for item in array
            val, indexes, isEvDependent = res
            return (mkArrayUpdater(val, indexes, invalidate), isEvDependent)
        else:
            # update for simple signal
            val, isEvDependent = res
            return (mkUpdater(val, invalidate), isEvDependent)

    def runCombProcesses(self):
        """
        Delta step for combinational processes
        """

        for proc in self.combProcsToRun:
            outContainer = self.outputContainers[proc]
            proc(self, outContainer)
            for actionSet in outContainer._all_signals:
                if actionSet:
                    res = self.conflictResolveStrategy(actionSet)
                    # prepare update
                    updater, isEvDependent = res
                    self.valuesToApply.append((actionSet.destination, updater, isEvDependent, proc))
                    actionSet.clear()
                # else value is latched

        self.combProcsToRun = UniqList()

    def runSeqProcesses(self, ev):
        """
        Delta step for sequential processes
        """
        updates = []
        for proc in self.seqProcsToRun:
            try:
                outContainer = self.outputContainers[proc]
            except KeyError:
                # processes does not have to have outputs
                outContainer = None

            proc(self, outContainer)

            if outContainer is not None:
                updates.append(outContainer)

        self.seqProcsToRun = UniqList()
        self.runSeqProcessesEv = None

        for cont in updates:
            for actionSet in cont._all_signals:
                if actionSet:
                    v = self.conflictResolveStrategy(actionSet)
                    updater, _ = v
                    actionSet.destination.simUpdateVal(self, updater)
                    actionSet.clear()

    def applyValues(self, ev):
        """
        Perform actual delta step
        """
        va = self.valuesToApply

        # log if there are items to log
        lav = self.config.logApplyingValues
        if va and lav:
            lav(self, va)
        # print(int(self.now // 1000), va)
        self.valuesToApply = []

        # apply values to signals, values can overwrite each other
        # but each signal should be driven by only one process and
        # it should resolve value collision
        addSp = self.seqProcsToRun.append
        for s, vUpdater, isEventDependent, comesFrom in va:
            if isEventDependent:
                # now=0 and this was process initialization or async reg
                addSp(comesFrom)
            else:
                # regular combinational process
                s.simUpdateVal(self, vUpdater)

        self.runCombProcesses()

        # processes triggered from simUpdateVal can add new values
        if self.valuesToApply:
            self.scheduleApplyValues()
            return

        # activate updateComplete if this was last applyValues() in this time
        self.updateComplete.succeed()  # trigger
        self.updateComplete = self.event()  # regenerate event
        self.applyValEv = None

    def read(self, sig):
        """
        Read value from signal or interface
        """
        try:
            v = sig._val
        except AttributeError:
            v = sig._sigInside._val

        return v.clone()

    def write(self, val, sig):
        """
        Write value to signal or interface.
        """
        # get target RtlSignal
        try:
            simSensProcs = sig.simSensProcs
        except AttributeError:
            sig = sig._sigInside
            simSensProcs = sig.simSensProcs

        # type cast of input value
        t = sig._dtype

        if isinstance(val, Value):
            v = val.clone()
            v = v._convert(t)
        else:
            v = t.fromPy(val)

        v.updateTime = self.now

        # can not update value in signal directly due singnal proxies
        sig.simUpdateVal(self, lambda curentV: (valueHasChanged(curentV, v), v))

        if not simSensProcs and self.applyValEv is not None:
            # in some cases simulation process can wait on all values applied
            # signal value was changed but there are no sensitive processes to it
            # because of this applyValues is never planed and but should be
            self.scheduleApplyValues()

    wait = BoundClass(Timeout)

    def simUnit(self, synthesisedUnit, time, extraProcesses=[]):
        """
        Run simulation
        """
        beforeSim = self.config.beforeSim
        if beforeSim is not None:
            beforeSim(self, synthesisedUnit)

        for p in extraProcesses:
            self.process(p(self))

        self._initUnitSignals(synthesisedUnit)

        self.run(until=time)




def seqEvalCond(cond):
    _cond = True
    for c in cond:
        _cond = _cond and bool(c.staticEval().val)

    return _cond


class CodeStatement():
    """
    Base class for code statements
    """

    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer, DebugTmpVarStack
        tmpVars = DebugTmpVarStack()
        ctx = VhdlSerializer.getBaseContext()
        ctx.createTmpVarFn = tmpVars.createTmpVarFn

        s = getattr(VhdlSerializer, self.__class__.__name__)(self, ctx)
        return "%s%s" % (tmpVars.serialize(), s)


class IfContainer(CodeStatement):
    """
    Structural container of if statement for hdl rendering
    """
    def __init__(self, cond, ifTrue=[], ifFalse=[], elIfs=[]):
        """
        :param cond: list of conditions for this if
        :param ifTrue: list of statements which should be active if cond. is met
        :param elIfs: list of tuples (list of conditions, list of statements)
        :param ifFalse: list of statements which should be active if cond.
            and any other cond. in elIfs is met
        """
        self.cond = cond
        self.ifTrue = ifTrue
        self.elIfs = elIfs
        self.ifFalse = ifFalse

    def seqEval(self):
        if seqEvalCond(self.cond):
            for s in self.ifTrue:
                s.seqEval()
        else:
            for c in self.elIfs:
                if seqEvalCond(c[0]):
                    for s in c[1]:
                        s.seqEval()
                    return

            for s in self.ifFalse:
                s.seqEval()


class SwitchContainer(CodeStatement):
    """
    Structural container for switch statement for hdl rendering
    """
    def __init__(self, switchOn, cases, default=[]):
        self.switchOn = switchOn
        self.cases = cases
        self.default = default

    def seqEval(self):
        raise NotImplementedError()


class WhileContainer(CodeStatement):
    """
    Structural container of while statement for hdl rendering
    """
    def __init__(self, cond, body):
        self.cond = cond
        self.body = body

    def seqEval(self):
        while seqEvalCond(self.cond):
            for s in self.body:
                s.seqEval()


class WaitStm(CodeStatement):
    """
    Structural container of wait statemnet for hdl rendering
    """
    def __init__(self, waitForWhat):
        self.isTimeWait = isinstance(waitForWhat, int)
        self.waitForWhat = waitForWhat



from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.defs import BIT, SLICE
from hwt.serializer.exceptions import UnsupportedEventOpErr
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


opPrecedence = {AllOps.NOT: 3,
                AllOps.NEG: 5,
                AllOps.RISING_EDGE: 0,
                AllOps.DIV: 6,
                AllOps.ADD: 7,
                AllOps.SUB: 7,
                AllOps.MUL: 3,
                AllOps.EQ: 10,
                AllOps.NEQ: 10,
                AllOps.AND: 11,
                AllOps.XOR: 12,
                AllOps.OR: 13,
                AllOps.DOWNTO: 2,
                AllOps.TO: 2,
                AllOps.GREATERTHAN: 9,
                AllOps.LOWERTHAN: 9,
                AllOps.GE: 9,
                AllOps.LE: 9,
                AllOps.CONCAT: 5,
                AllOps.INDEX: 1,
                AllOps.TERNARY: 16,
                AllOps.CALL: 2,
                # AllOps.SHIFTL:8,
                # AllOps.SHIFTR:8,
                }


class SystemCSerializer_ops():
    @classmethod
    def Operator(cls, op, ctx):
        def p(operand):
            s = cls.asHdl(operand, ctx)
            if isinstance(operand, RtlSignalBase):
                try:
                    o = operand.singleDriver()
                    if (o.operator != op.operator and
                            opPrecedence[o.operator] <= opPrecedence[op.operator]):
                        return "(%s)" % s
                except Exception:
                    pass
            return s

        ops = op.operands
        o = op.operator

        def _bin(name):
            return name.join(map(p, ops))

        if o == AllOps.AND:
            return _bin(' & ')
        elif o == AllOps.OR:
            return _bin(' | ')
        elif o == AllOps.XOR:
            return _bin(' ^ ')
        elif o == AllOps.NOT:
            assert len(ops) == 1
            return "~" + p(ops[0])
        elif o == AllOps.CALL:
            return "%s(%s)" % (cls.FunctionContainer(ops[0]), ", ".join(map(p, ops[1:])))
        elif o == AllOps.CONCAT:
            return _bin(' & ')
        elif o == AllOps.DIV:
            return _bin(' / ')
        # elif o == AllOps.DOWNTO:
        #    return _bin(':')
        # elif o == AllOps.TO:
        #    return _bin(':')
        elif o == AllOps.EQ:
            return _bin(' == ')
        elif o == AllOps.NEQ:
            return _bin(' != ')
        elif o == AllOps.GREATERTHAN:
            return _bin(' > ')
        elif o == AllOps.LOWERTHAN:
            return _bin(' < ')
        elif o == AllOps.GE:
            return _bin(' >= ')
        elif o == AllOps.LE:
            return _bin(' <= ')
        elif o == AllOps.INDEX:
            assert len(ops) == 2
            o0, o1 = ops
            o0_str = cls.asHdl(o0, ctx)
            if ops[1]._dtype == SLICE:
                return "%s.range(%s, %s)" % (o0_str, p(o1.val[0]), p(o1.val[1]))
            else:
                return "%s[%s]" % (o0_str, p(o1))

        elif o == AllOps.SUB:
            return _bin(' - ')
        elif o == AllOps.MUL:
            return _bin(' * ')
        elif o == AllOps.ADD:
            return _bin(' + ')
        elif o == AllOps.TERNARY:
            zero, one = BIT.fromPy(0), BIT.fromPy(1)
            if ops[1] == one and ops[2] == zero:
                # ignore redundant x ? 1 : 0
                return cls.condAsHdl([ops[0]], True, ctx)
            else:
                return "%s ? %s : %s" % (cls.condAsHdl([ops[0]], True, ctx),
                                         p(ops[1]),
                                         p(ops[2]))
        elif o == AllOps.RISING_EDGE or o == AllOps.FALLIGN_EDGE:
            if ctx.isSensitivityList:
                if o == AllOps.RISING_EDGE:
                    _o = ".pos()"
                else:
                    _o = ".neg()"
                return p(ops[0]) + _o
            else:
                raise UnsupportedEventOpErr()
        elif o == AllOps.BitsToInt:
            return p(ops[0])
        elif o in [AllOps.BitsAsSigned, AllOps.BitsAsUnsigned, AllOps.BitsAsVec, AllOps.IntToBits]:
            assert len(ops) == 1
            return "static_cast<%s>(%s)" % (cls.HdlType(op.result._dtype, ctx), p(ops[0]))
        elif o == AllOps.POW:
            assert len(ops) == 2
            raise NotImplementedError()
            # return _bin('**')
        else:
            raise NotImplementedError("Do not know how to convert %s to vhdl" % (o))



from hwt.hdlObjects.types.boolean import Boolean
from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.string import String
from hwt.hdlObjects.types.slice import Slice

"""
Definitions of most common types
"""

BOOL = Boolean()
INT = Integer()
BIT = Bits(width=1)
BIT_N = Bits(width=1, negated=True)
STR = String()
SLICE = Slice()



from hwt.hdlObjects.types.bits import Bits


class SystemCSerializer_type():
    @classmethod
    def HdlType_bits(cls, typ, ctx, declaration=False):
        if declaration:
            raise NotImplementedError()
        
        w = typ.bit_length()
        
        if w <= 64:
            if typ.signed:
                typeBaseName = "int" 
            else:
                typeBaseName = "uint"
        else:
            if typ.signed:
                typeBaseName = "bigint" 
            else:
                typeBaseName = "biguint"
            
        return "sc_%s<%d>" % (typeBaseName, w)
    
    @classmethod
    def HdlType_enum(cls, typ, ctx, declaration=False):
        if declaration:
            raise TypeError("There is problem with tracing of c enums, use Bits instead")
        else:
            valueCnt = len(typ._allValues)
            return cls.HdlType_bits(Bits(valueCnt.bit_length()), ctx, declaration=declaration)


from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.booleanVal import BooleanVal
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.hdlObjects.value import Value


def convertInteger(self, sigOrVal, toType):
    isVal = isinstance(sigOrVal, Value)
    if toType == BOOL:
        if isVal:
            v = sigOrVal.val
            assert v == 0 or v == 1
            return BooleanVal(v, BOOL, sigOrVal.vldMask, sigOrVal.updateTime)

    elif isinstance(toType, Bits):
        if isVal:
            _v = sigOrVal.val
            w = toType.bit_length()
            assert _v.bit_length() <= w, "%d can not fit into %d bits" % (_v, w)
            v = toType.fromPy(_v)

            v.updateTime = sigOrVal.updateTime

            v._dtype = toType
            if not sigOrVal.vldMask:
                v.vldMask = 0
            return v
        else:
            return Operator.withRes(AllOps.IntToBits, [sigOrVal], toType)

    return HdlType.defaultConvert(self, sigOrVal, toType)
